@@ -13,8 +13,8 @@
 
 <par>It contains different &sax;2 parser driver classes (mostly for sgmlop, everything else
 is from <app moreinfo="http://pyxml.sf.net/">PyXML</app>). It includes a
-<pyref class="HTMLParser"><class>HTMLParser</class></pyref> that uses sgmlop to parse &html;
-and emit &sax;2 events.</par>
+<pyref class="HTMLParser"><class>HTMLParser</class></pyref> that uses sgmlop
+to parse &html; and emit &sax;2 events.</par>
 """
 
 import sys, os, os.path, warnings, cStringIO
@@ -98,12 +98,7 @@ class SGMLOPParser(sax.xmlreader.XMLReader, sax.xmlreader.Locator):
 					self.lineNumber += 1
 				self.feed(data)
 				self.parsed = True
-		except SystemExit:
-			self.close()
-			self.source = None
-			self.encoding = None
-			raise
-		except KeyboardInterrupt:
+		except (SystemExit, KeyboardInterrupt):
 			self.close()
 			self.source = None
 			self.encoding = None
@@ -229,7 +224,7 @@ class SGMLOPParser(sax.xmlreader.XMLReader, sax.xmlreader.Locator):
 		"""
 		if text is None:
 			return xsc.Null
-		ct = self.getContentHandler().createText
+		ct = getattr(self.getContentHandler(), "createText", xsc.Text)
 		node = xsc.Frag()
 		while True:
 			texts = text.split(u"&", 1)
@@ -449,31 +444,32 @@ class Parser(object):
 		&xist; itself provides several &sax;2 parsers
 		(all based on Fredrik Lundh's <app>sgmlop</app> from <app moreinfo="http://pyxml.sf.net/">PyXML</app>):</par>
 		<ulist>
-		<item><pyref module="ll.xist.parsers" class="SGMLOPParser"><class>ll.xist.parsers.SGMLOPParser</class></pyref>
-		(which is the default if the <arg>parser</arg> argument is not given);</item>
-		<item><pyref module="ll.xist.parsers" class="BadEntityParser"><class>ll.xist.parsers.BadEntityParser</class></pyref>
+		<item><pyref module="ll.xist.parsers" class="SGMLOPParser"><class>SGMLOPParser</class></pyref>
+		(which is the default if the <arg>saxparser</arg> argument is not given);</item>
+		<item><pyref module="ll.xist.parsers" class="BadEntityParser"><class>BadEntityParser</class></pyref>
 		(which is based on <class>SGMLOPParser</class> and tries to pass on unknown entity references as literal content);</item>
 		<item><pyref module="ll.xist.parsers" class="HTMLParser"><class>HTMLParser</class></pyref> (which is
 		based on BadEntityParser and tries to make sense of &html; sources).</item>
 		</ulist>
 		</item>
-	
+
+		<term><arg>tidy</arg></term><item>If <arg>tidy</arg> is true, <link href="http://xmlsoft.org/">libxml2</link>'s
+		&html; parser will be used for parsing broken &html;.</item>
 		<term><arg>nspool</arg></term><item>an instance of <pyref module="ll.xist.xsc" class="NSPool"><class>ll.xist.xsc.NSPool</class></pyref>;
 		From this namespace pool namespaces will be taken when the parser
 		encounters <lit>xmlns</lit> attributes.</item>
 
 		<term><arg>prefixes</arg></term><item>an instance of <pyref module="ll.xist.xsc" class="Prefixes"><class>ll.xist.xsc.Prefixes</class></pyref>;
 		Specifies which namespace modules should be available during parsing
-		and to which prefixes they are mapped (but of course this
-		mapping can be changed during parsing by using <lit>xmlns</lit>
-		attributes in the usual way).</item>
+		and to which prefixes they are mapped (this mapping can change during
+		parsing by <lit>xmlns</lit> attributes are encountered).</item>
 
 		<term><arg>loc</arg></term><item>Should location information be attached to the generated nodes?</item>
 
 		<term><arg>validate</arg></term><item>Should the parsed &xml; nodes be validated after parsing?</item>
 
 		<term><arg>encoding</arg></term><item>The default encoding to use, when the
-		source doesn't provide an &xml; header. The default <lit>None</lit> results in
+		source doesn't provide an encoding. The default <lit>None</lit> results in
 		<lit>"utf-8"</lit> for parsing &xml; and <lit>"iso-8859-1"</lit> when parsing
 		broken &html; (when <lit><arg>tidy</arg></lit> is true).</item>
 		</dlist>
@@ -496,8 +492,8 @@ class Parser(object):
 
 	def _last(self):
 		"""
-		return the newest node from the stack that is a real node.
-		(There might be false node on the stack, because we are inside
+		Return the newest node from the stack that is a real node.
+		(There might be fake nodes on the stack, because we are inside
 		of illegal elements).
 		"""
 		for (node, prefixes) in reversed(self._nesting):
@@ -505,6 +501,9 @@ class Parser(object):
 				return node
 
 	def _parseHTML(self, stream, base, sysid, encoding):
+		"""
+		Internal helper method for parsing &html; via <module>libxml2</module>.
+		"""
 		import libxml2 # This requires libxml2 (see http://www.xmlsoft.org/)
 
 		def decode(s):
@@ -513,9 +512,6 @@ class Parser(object):
 			except UnicodeDecodeError:
 				return s.decode("iso-8859-1")
 
-		data = stream.read()
-		doc = libxml2.htmlReadMemory(data, len(data), sysid, encoding, 0x60)
-		ns = self.nspool.get(html.xmlname, html)
 		def toxsc(node):
 			if node.type == "document_html":
 				newnode = xsc.Frag()
@@ -527,6 +523,8 @@ class Parser(object):
 				name = decode(node.name).lower()
 				try:
 					newnode = ns.element(name, xml=True)()
+					if self.loc:
+						newnode.startloc = xsc.Location(sysid=sysid, line=node.lineNo())
 				except xsc.IllegalElementError:
 					newnode = xsc.Frag()
 				else:
@@ -554,14 +552,27 @@ class Parser(object):
 					newnode.parsed(self, start=False)
 			elif node.type in ("text", "cdata"):
 				newnode = xsc.Text(decode(node.content))
+				if self.loc:
+					newnode.startloc = xsc.Location(sysid=sysid, line=node.lineNo())
 			elif node.type == "comment":
 				newnode = xsc.Comment(decode(node.content))
+				if self.loc:
+					newnode.startloc = xsc.Location(sysid=sysid, line=node.lineNo())
 			else:
 				newnode = xsc.Null
-			newnode.startloc = xsc.Location(sysid=sysid) # FIXME: get line from node.lineNo()
 			return newnode
-		node = toxsc(doc)
-		doc.freeDoc()
+
+		data = stream.read()
+		ns = self.nspool.get(html.xmlname, html)
+		try:
+			olddefault = libxml2.lineNumbersDefault(1)
+			doc = libxml2.htmlReadMemory(data, len(data), sysid, encoding, 0x160)
+			try:
+				node = toxsc(doc)
+			finally:
+				doc.freeDoc()
+		finally:
+			libxml2.lineNumbersDefault(olddefault)
 		return node
 
 	def _parse(self, stream, base, sysid, encoding):
@@ -613,17 +624,27 @@ class Parser(object):
 		return root
 
 	def parse(self, stream, base=None, sysid=None):
+		"""
+		Parse &xml; from the stream <arg>stream</arg> into an &xist; tree.
+		<arg>base</arg> is the base &url; for the parsing process, <arg>sysid</arg>
+		is the &xml; system identifier (defaulting to <arg>base</arg> if it is <lit>None</lit>).
+		"""
 		if sysid is None:
 			sysid = base
 		return self._parse(stream, base, sysid, self.encoding)
 
-	def parseString(self, string, base=None, sysid=None):
-		if isinstance(string, unicode):
+	def parseString(self, text, base=None, sysid=None):
+		"""
+		Parse the string <arg>text</arg> (<class>str</class> or <class>unicode</class>)
+		into an &xist; tree. <arg>base</arg> is the base &url; for the parsing process, <arg>sysid</arg>
+		is the &xml; system identifier (defaulting to <arg>base</arg> if it is <lit>None</lit>).
+		"""
+		if isinstance(text, unicode):
 			encoding = "utf-8"
-			string = string.encode(encoding)
+			text = text.encode(encoding)
 		else:
 			encoding = self.encoding
-		stream = cStringIO.StringIO(string)
+		stream = cStringIO.StringIO(text)
 		if base is None:
 			base = url.URL("STRING")
 		if sysid is None:
@@ -631,6 +652,17 @@ class Parser(object):
 		return self._parse(stream, base, sysid, encoding)
 
 	def parseURL(self, name, base=None, sysid=None, headers=None, data=None):
+		"""
+		Parse &xml; input from the &url; <arg>name</arg> which might be a string
+		or an <pyref module="ll.url" class="URL"><class>URL</class></pyref> object
+		into an &xist; tree. <arg>base</arg> is the base &url; for the parsing process
+		(defaulting to the final &url; of the response (i.e. including redirects)),
+		<arg>sysid</arg> is the &xml; system identifier (defaulting to <arg>base</arg>
+		if it is <lit>None</lit>). <arg>headers</arg> is a dictionary with additional
+		headers used in the &http; request. If <arg>data</arg> is not <lit>None</lit>
+		it must be a dictionary. In this case <arg>data</arg> will be used as the data
+		for a <lit>POST</lit> request.
+		"""
 		name = url.URL(name)
 		stream = name.openread(headers=headers, data=data)
 		if base is None:
@@ -642,11 +674,16 @@ class Parser(object):
 			encoding = stream.encoding
 		return self._parse(stream, base, sysid, encoding)
 
-	def parseFile(self, name, base=None, sysid=None):
-		name = os.path.expanduser(name)
-		stream = open(name, "rb")
+	def parseFile(self, filename, base=None, sysid=None):
+		"""
+		Parse &xml; input from the file named <arg>filename</arg>. <arg>base</arg> is
+		the base &url; for the parsing process (defaulting to <arg>filename</arg>),
+		<arg>sysid</arg> is the &xml; system identifier (defaulting to <arg>base</arg>).
+		"""
+		filename = os.path.expanduser(filename)
+		stream = open(filename, "rb")
 		if base is None:
-			base = url.File(name)
+			base = url.File(filename)
 		if sysid is None:
 			sysid = str(base)
 		return self._parse(stream, base, sysid, self.encoding)
@@ -661,31 +698,35 @@ class Parser(object):
 		pass
 
 	def startElement(self, name, attrs):
+		oldprefixes = self.prefixes
+
 		newprefixes = {}
 		for (attrname, attrvalue) in attrs.items():
 			if attrname==u"xmlns" or attrname.startswith(u"xmlns:"):
 				ns = self.nspool[unicode(attrvalue)]
 				newprefixes[attrname[6:] or None] = [ns]
 
-		prefixes = self.prefixes
 		if newprefixes:
-			prefixes = prefixes.clone()
+			prefixes = oldprefixes.clone()
 			prefixes.update(newprefixes)
-			self.prefixes = prefixes
+			self.prefixes = newprefixes = prefixes
+		else:
+			newprefixes = oldprefixes
+
 		node = self.createElement(name)
 		if node is not None:
 			for (attrname, attrvalue) in attrs.items():
 				if attrname != u"xmlns" and not attrname.startswith(u"xmlns:"):
-					attrname = prefixes.attrnamefromqname(node, attrname)
+					attrname = newprefixes.attrnamefromqname(node, attrname)
 					if attrname is not None: # None means an illegal attribute
 						node[attrname] = attrvalue
 						node[attrname].parsed(self)
 			node.attrs.parsed(self)
 			node.parsed(self, start=True)
 			self.__appendNode(node)
-		# push new innermost element onto the stack, together with the list of prefix mappings defined by this node
+		# push new innermost element onto the stack, together with the list of prefix mappings to which we have to return when we leave this element
 		# If the element is bad (i.e. createElement returned None), we push None as the node
-		self._nesting.append((node, prefixes))
+		self._nesting.append((node, oldprefixes))
 		self.skippingwhitespace = False
 
 	def endElement(self, name):
@@ -696,12 +737,14 @@ class Parser(object):
 				currentelement.checkvalid()
 			if self.loc:
 				currentelement.endloc = self.getLocation()
-		self.prefixes = self._nesting.pop()[1] # pop the innermost element off the stack and restore the old prefixes mapping
 
-		# We have to check after the old prefix mapping from this element has been dropped
+		# We have to check with the currently active prefixes
 		element = self.createElement(name) # Unfortunately this creates the element a second time.
 		if element is not None and element.__class__ is not currentelement.__class__:
 			raise xsc.ElementNestingError(currentelement.__class__, element.__class__)
+
+		self.prefixes = self._nesting.pop()[1] # pop the innermost element off the stack and restore the old prefixes mapping (from outside this element)
+
 		self.skippingwhitespace = False
 
 	def characters(self, content):
@@ -803,21 +846,55 @@ class Parser(object):
 		return entity
 
 
-def parse(stream, base, sysid=None, **parserargs):
+def parse(stream, base=None, sysid=None, **parserargs):
+	"""
+	Parse &xml; from the stream <arg>stream</arg> into an &xist; tree.
+	For the arguments <arg>base</arg> and <arg>sysid</arg> see the method
+	<pyref class="Parser" method="parse"><method>parse</method></pyref>
+	in the <class>Parser</class> class. You can pass any other argument that the
+	<pyref class="Parser" method="__init__"><class>Parser</class> constructor</pyref>
+	takes as keyword arguments via <arg>parserargs</arg>.
+	"""
 	parser = Parser(**parserargs)
 	return parser.parse(stream, base, sysid)
 
 
 def parseString(text, base=None, sysid=None, **parserargs):
+	"""
+	Parse the string <arg>text</arg> (<class>str</class> or <class>unicode</class>) into an
+	&xist; tree. For the arguments <arg>base</arg> and <arg>sysid</arg> see the method
+	<pyref class="Parser" method="parseString"><method>parseString</method></pyref>
+	in the <class>Parser</class> class. You can pass any other argument that the
+	<pyref class="Parser" method="__init__"><class>Parser</class> constructor</pyref>
+	takes as keyword arguments via <arg>parserargs</arg>.
+	"""
 	parser = Parser(**parserargs)
 	return parser.parseString(text, base, sysid)
 
 
 def parseURL(url, base=None, sysid=None, headers=None, data=None, **parserargs):
+	"""
+	Parse &xml; input from the &url; <arg>name</arg> which might be a string
+	or an <pyref module="ll.url" class="URL"><class>URL</class></pyref> object
+	into an &xist; tree. For the arguments <arg>base</arg>, <arg>sysid</arg>,
+	<arg>headers</arg> and <arg>data</arg> see the method
+	<pyref class="Parser" method="parseURL"><method>parseURL</method></pyref>
+	in the <class>Parser</class> class. You can pass any other argument that the
+	<pyref class="Parser" method="__init__"><class>Parser</class> constructor</pyref>
+	takes as keyword arguments via <arg>parserargs</arg>.
+	"""
 	parser = Parser(**parserargs)
 	return parser.parseURL(url, base, sysid, headers, data)
 
 
 def parseFile(filename, base=None, sysid=None, **parserargs):
+	"""
+	Parse &xml; input from the file named <arg>filename</arg>. For the arguments
+	<arg>base</arg> and <arg>sysid</arg> see the method
+	<pyref class="Parser" method="parseFile"><method>parseFile</method></pyref>
+	in the <class>Parser</class> class. You can pass any other argument that the
+	<pyref class="Parser" method="__init__"><class>Parser</class> constructor</pyref>
+	takes as keyword arguments via <arg>parserargs</arg>.
+	"""
 	parser = Parser(**parserargs)
 	return parser.parseFile(filename, base, sysid)
